#                          InuBBS version 0.5
#                                inubbs.pl
#                   Copyright (C) 2002-2009 Dai Sato
#                            satodai@w3m.jp

### Main Routine
sub inubbs() {
  if( $USE_COPY ){
    use File::Copy;
  }

  $VERSION = 'Version 0.5';

  local( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst )
    = localtime( time );
  local( $mname )
    = ( 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
       'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec' )[$mon];
  local( $dname ) = ( 'Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat' )[$wday];
  local( $y ) = $year + 1900;
  local( $m ) = ( '01' .. '12' )[$mon];
  local( $d ) = ( '00' .. '31' )[$mday];
  local( $hh ) = ( '00' .. '23' )[$hour];
  local( $mm ) = ( '00' .. '59' )[$min];
  local( $ss ) = ( '00' .. '59' )[$sec];

  $TAGLIST = ',p,br,blockquote,pre,hr,h1,h2,h3,h4,h5,h6,ul,ol,dl,li,dt,dd,strong,em,cite,del,tt,a,b,i,u,strike,big,small,font,sub,sup,div,span,';
  #---------------------------------------------------------------------------

  # HTTP header
  print "Content-type: text/html\n\n";

  # Get Variables
  ReadParse( );

  while( ( $key, $value ) = each %in ){
    $$key = $value;
  }
  ConvertMsgVarsTo( 'euc' );

  local( $REAL_LOG, $reallog );
  if( $USE_REV ){
    local( $REAL_REV, $realrev );
  }

  if( @ARGV[ 0 ] eq 'digest' || $digest ne "" ){
    HTMLdigest( $digest );
  } elsif( @ARGV[ 0 ] eq 'search' || $search ne "" ){
    Search( $search );
  } elsif( $post || $posttext ) {
    CreateLogFile( );
    PostMessage( );
  } else {
    InputMessage( );
  }
}

#####
# Lock & Unlock ###

sub Lock {
  local( $filename ) = @_;
  local( $LOCK_EX ) = 2;

  open( LOCK, ">$filename" );
  flock( LOCK, $LOCK_EX );
  close( LOCK );
}

# Unlock
sub Unlock {
  local( $filename ) = @_;
  local( $LOCK_UN ) = 8;

  open( LOCK, ">$filename" );
  flock( LOCK, $LOCK_UN );
  close( LOCK );
}

#####
# Convert Message Variables to $KANJI_CODE
sub ConvertMsgVarsTo {
  local( $char_code ) = @_;

  &jcode'convert( *TITLE, $char_code );
  &jcode'convert( *DOGTAG, $char_code );
  &jcode'convert( *MAINTAINER, $char_code );
  &jcode'convert( *POSTMSG, $char_code );
  &jcode'convert( *READNEW, $char_code );
  &jcode'convert( *READREV, $char_code );
  &jcode'convert( *READOLD, $char_code );
  &jcode'convert( *READRULE, $char_code );
  &jcode'convert( *GOLINKS, $char_code );
  &jcode'convert( *GOHOME, $char_code );
  &jcode'convert( *GOTOTOP, $char_code );
  for( $i = 0; $LINKS[ $i ]; $i++ ){
    &jcode'convert( $LINKS[ $i ], $char_code );
  }

  &jcode'convert( *author, $char_code, $KANJI_CODE, "z" );
  &jcode'convert( *mail_address, $char_code, $KANJI_CODE, "z" );
  &jcode'convert( *subject, $char_code, $KANJI_CODE, "z" );
  &jcode'convert( *message, $char_code, $KANJI_CODE, "z" );
}

#####
# Check tags
sub CheckTag {
  local( $msg ) = @_;
  local( $tmp);

  $_ = $msg;
  for($i = 0; m/<[ \s]*\/?(\w*)/ig && $i < 50; $i++ ){
    $tmp = $1;
    if( !( $TAGLIST =~ /,$tmp,/i) ){
      return $tmp;
    }
    else{
    }
  }
  return;
}

#####
# make Message-ID

sub MakeMessageID {
  $msgid = "$y$m$d$hh$mm$ss\@$DOGTAG"; # global variable

  return $msgid;
}

#####
# convert "http://..." to A HREF tag

sub URL2tag {
  local( $tmpstr ) = @_;

  $tmpstr =~ s/(https?:\/\/[a-zA-Z0-9;\/\?:@&=+$,---_.!~*'()<>#%"{}|\\^\[\]`]+)/<a href=\"$1\">$1<\/a>/g;
  return $tmpstr;
}

#####
# HTML escape

sub HTMLescape {
  local( $tmpstr ) = @_;

  &jcode'convert( *tmpstr, 'euc' );
  $tmpstr =~ s/&/&amp;/g;
  $tmpstr =~ s/</&lt;/g;
  $tmpstr =~ s/>/&gt;/g;
  $tmpstr =~ s/\"/&quot;/g;
  return $tmpstr;
}

#####
# HTML head

sub HTMLhead {
  local( $title ) = @_;
  local( $tmp ) = "";

  $tmp .= "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n";
  $tmp .= "<html lang=\"ja\" dir=\"ltr\"><head>\n";
  $tmp .= "<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; charset=$CHARSET\">\n";
  $tmp .= "<META HTTP-EQUIV=\"CONTENT-STYLE-TYPE\" CONTENT=\"text/css\">\n";
  $tmp .= "<link rev=\"made\" href=\"mailto:$MAILADD\">\n";
  $tmp .= "<LINK rel=\"INDEX\" href=\"$HOMEPAGE\">\n";
  $tmp .= "<link rel=\"help\" href=\"../$RULE_HTML\">\n";
  $tmp .= "<link rel=\"contents\" href=\"./$OLD_HTML\">\n";
  if( $FAVICON ){
    $tmp .= "<LINK REL=\"icon\" type=\"image/gif\" href=\"../$FAVICON\">\n";
  }
  if( $STYLE_FILE ){
    $tmp .= '<LINK REL="StyleSheet" HREF="';
    $tmp .= $STYLE_FILE;
    $tmp .= '" TYPE="text/css">'."\n";
  }
  if( $STYLE ){
    $tmp .= "<style>\n$STYLE\n</style>\n";
  }
  if( $USE_FORMULA ){
    $tmp .= <<END;
<script><!--
function getFmlElmtId( id, num ){
 if( num == 0 ){
  var tmpStr = id + '-body';
 }else{
  var tmpStr = 'fml' + id + '-' + num;
 }
 return tmpStr;
}

function mkIfrVisible( ifr, num ){
 // make iframe visible
 ifr.style.width = '100%';
 if( num == 0 ){
  ifr.style.height = '200px';
 }else{
  ifr.style.height = '70px';
 }
}

function closeFormula( id, num ){
 var fmlElmtId = getFmlElmtId( id, num );
 var fmlElmt = document.getElementById( fmlElmtId );
 if( num == 0 ){
  var obj = document.getElementById( id + '-button' );
  obj.innerHTML = obj.innerHTML.replace( /closeArticle/g,'showArticle');
  var targetText = fmlElmt.innerHTML;
 }else{
  var targetText = fmlElmt.innerHTML.replace( /closeFormula/g,'showFormula');
  fmlElmt.innerHTML = targetText;
 }
 fmlElmt.removeChild( document.getElementById( 'ifr-' + fmlElmtId ) );
}

function showFormula( id, num ){
 // get the ID of the object
 var fmlElmtId = getFmlElmtId( id, num );
 // get the formula-SPAN or DIV and rewrite show/close buttons
 var fmlElmt = document.getElementById( fmlElmtId );
 if( num == 0 ){
  var obj = document.getElementById( id + '-button' );
  obj.innerHTML = obj.innerHTML.replace( /showArticle/g,'closeArticle');
  var targetText = fmlElmt.innerHTML;
 }else{
  var targetText = fmlElmt.innerHTML.replace( /showFormula/g,'closeFormula');
  fmlElmt.innerHTML = targetText;
 }

 // constructing the content for iframe
 var formulaContent;
 formulaContent = '<html><head><title>formula: ' + id + '</title>';
 formulaContent += '<script type="text/javascript" src="$USE_FORMULA"></script>';
 formulaContent += '</head><body bgcolor="#FFFFFF">';
 formulaContent += targetText.replace( /<a class="button" [^<]*<\\/a>/g, '' ).replace( /<iframe .*<\\/iframe>/g, '' );
 formulaContent += '</body></html>';

 // creating an iframe
 var ifr = document.createElement('iframe');
 ifr.id = 'ifr-' + fmlElmtId;
 ifr.style.border = 'none';
 ifr.style.width = '0';
 ifr.style.height = '0';
 fmlElmt.appendChild( ifr );

 // loading the content into iframe
 var w = ifr.contentWindow;
 w.document.open( "text/html" );
 w.document.write( formulaContent );
 w.document.close();
 mkIfrVisible( ifr, num );
}

function closeArticleInMathML( id, num ){
 closeFormula( id, num );
}

function showArticleInMathML( id, num ){
 showFormula( id, num );
}
// --></script>
END
  }
  $tmp .= "<title>$title</title>\n";
  $tmp .= "</head>\n\n";
  $tmp .= "<$BODY>\n\n";
  if( $LOGO ){
    $tmp .= "$LOGO\n\n<h1 class=\"title\"><a name=\"top\">$title</a></h1>\n\n<div class=\"body\">\n\n";
  } else {
    $tmp .= "<h1 class=\"title\"><a name=\"top\">$title</a></h1>\n\n";
  }

  &jcode'convert( *tmp, 'euc' );
  return $tmp;
}

#####
# HTML foot

sub HTMLfoot {
  local( $tmp );

  $tmp .= "<hr>\n\n";
  $tmp .= "<div>$MAINTAINER</div>\n\n";
  $tmp .= "<p style=\"text-align:right\">";
  $tmp .= "<a href=\"http://dog.w3m.jp/software/inubbs/\">InuBBS</a> $VERSION</p>\n";
  if( $LOGO ){
    $tmp .= "</div>";
  }
  $tmp .= "</body></html>\n";

  &jcode'convert( *tmp, 'euc' );
  return( $tmp );
}

#####
# Make HTML links

sub HTMLlinks {
  local( $dir ) = @_;
  local( $tmp, $spl ) = "";

  $tmp .= "<p>\n";

  if( $dir eq 'spool' ){
    $tmp .= "<a href=\"../$BBS_CGI\">$POSTMSG</a>\n";
    $spl = '';
    $up  = '../';
  }else{
    $spl = "$SPOOL/";
    $up  = '';
  }
  $tmp .= "<a href=\"$spl$LOG_HTML";
  if( $dir ne 'spool' && ( $post || $posttext ) ) {
    $tmp .= "?$y$m$d$hh$mm$ss";
  }
  $tmp .= "\">$READNEW</a>\n";
  if( $USE_REV ){
    $tmp .= "<a href=\"$spl$LOG_REV_HTML\">$READREV</a>\n";
  }
  if( -r "$TOPDIR/$SPOOL/$OLD_HTML" ) {
    $tmp .= "<a href=\"$spl$OLD_HTML\">$READOLD</a>\n";
  }
  if( -r "$TOPDIR/$RULE_HTML" ) {
    $tmp .= "<a href=\"$up$RULE_HTML\">$READRULE</a>\n";
  }
  if( -r "$TOPDIR/$LINKS_HTML" ) {
    $tmp .= "<a href=\"$up$LINKS_HTML\">$GOLINKS</a>\n";
  }
  $tmp .= "<a href=\"$HOMEPAGE\">$GOHOME</a>\n";
  if( $dir eq 'spool' ){
    for( $i = 0; $LINKS[ $i ]; $i++ ){
      if( $i == 0 ){
        $tmp .= "<br>\n";
      }
      $tmp .= "<a href=\"$LINKS_URL[ $i ]\">$LINKS[ $i ]</a>\n";
    }
  }
  $tmp .= "</p>\n";
  $tmp .= "\n";

  &jcode'convert( *tmp, 'euc' );
  return $tmp;
}

#####
# make input file for weblint

sub MakeWLtmp{
  local( $file ) = @_;

  if( open( WLIN, ">$wlin" )) {
    $msg = ParseMessage();

    print WLIN HTMLhead( "Your Message" );
    print WLIN "$msg\n";
    print WLIN "</body></html>\n";
    close( WLIN );
  }
}

#####
# Result of weblint

sub ResultWeblint {
  local( $wlin ) = "$TOPDIR/$SPOOL/wlin$$.html";
  local( $wlout ) = "$TOPDIR/$SPOOL/wlout$$.html";
  local( $tmp ) = "";
  local( $tmpwl ) = "";

  if( !MakeWLtmp( $wlin ) ){
    $tmp = "<p>Can not make tmp file for weblint.</p>\n\n";
    return $tmp;
  }

  $tmp .= "<hr>\n";
  $tmp .= "\n";
  $tmp .= "<h2>Weblint</h2>\n";
  $tmp .= "\n";

  local( $cmd ) = "$WEBLINT $wlin > $wlout";
  local( $status ) = system( $cmd );
  unlink $wlin;

  if( ! open( WLOUT, "$wlout" ) ){
    $tmp .= "<p>Can not read the result of weblint.</p>\n\n";
    unlink $wlout;
    return $tmp;
  }

  $_ = <WLOUT>;
  if( $_ eq '' ){
    if( $status ){
      $cmd = HTMLescape( $cmd );
      $tmp .= "<p>Can not execute \"$cmd\".<br>$status</p>\n";
    }else{
      $tmp .= "<p>Successful completion.</p>\n";
    }
    $tmp .= "\n";
  }else{
    $tmp .= "<pre>";

    $tmpwl .= $_;
    while( <WLOUT> ){
      $tmpwl .= $_;
    }
    $tmpwl = HTMLescape( $tmpwl );
    $tmpwl =~ s/^$wlin//g;
    &jcode'convert( *tmpwl, 'euc' );
    $tmp .= $tmpwl;
    $tmp .= "</pre>\n\n";
    $wlerror = 1;
  }
  close( WLOUT );
  unlink $wlout;

  return $tmp;
}

#####
# Input Form

sub PrintInputForm() {
  local( $tmp ) = "";
  local( $tmptxt ) = "";

  $tmp = HTMLescape( $author );
  $tmptxt .= "<p>\n";
  $tmptxt .= "Name: <br>\n";
  $tmptxt .= "<input type=\"text\" size=\"25\" name=\"author\" value=\"$tmp\">\n";

  if( $NEED_NON_ASCII ){
    if( $esp ){ $tmp = "CHECKED" } else { $tmp = "" }
    $tmptxt .= "<input type=\"checkbox\" name=\"esp\" $tmp >\n";
    $tmptxt .= $NEED_NON_ASCII;
    $tmptxt .= "\n</p>\n\n";
  }

  $tmp = HTMLescape( $mail_address );
  $tmptxt .= "<p>\n";
  $tmptxt .= "E-mail Address or URL: <br>\n";
  $tmptxt .= "<input type=\"text\" size=\"50\" name=\"mail_address\" value=\"$tmp\">\n";
  $tmptxt .= "</p>\n\n";

  $tmp = HTMLescape( $subject );
  $tmptxt .= "<p>\n";
  $tmptxt .= "Subject: <br>\n";
  $tmptxt .= "<input type=\"text\" size=\"50\" name=\"subject\" value=\"$tmp\">\n";
  $tmptxt .= "</p>\n\n";

  $tmptxt .= "<p>\n";
  $tmptxt .= "Message: <br>\n";

  $tmp = HTMLescape( $message );
  $tmptxt .= "<textarea rows=20 cols=70 name=\"message\">$tmp</textarea><br>\n";
  $tmptxt .= "</p>\n";
  $tmptxt .= "\n";

  $tmptxt .= "<p>\n";
  $tmptxt .= "<input type=\"submit\" value=\"Preview HTML\" name=\"preview\">\n";
  $tmptxt .= "<input type=\"submit\" value=\"Preview TEXT\" name=\"previewtext\">\n";
  $tmptxt .= "<input type=\"reset\" value=\"Reset\">\n";
  $tmptxt .= "</p>\n\n";

  $tmptxt .= "</form>\n\n";

  jcode'convert( *tmptxt, $KANJI_CODE );
  print $tmptxt;
}

#####
# construct tags to show formulas

sub mkFormulaTag {
  local( $id, $n, $org ) = @_;
  local( $form_tag );

  $form_tag = '<span id="fml';
  $form_tag .= "$id-$n";
  $form_tag .= '">';
  $form_tag .= "&#96;$org&#96;";
  $form_tag .= '<a class="button" onClick="javascript:showFormula(';
  $form_tag .= "\'$msgid\', \'$n\'";
  $form_tag .= ')">showFormula</a>';
  $form_tag .= '</span>';
  return $form_tag;
}

#####
# Parse message

sub ParseMessage {
  local( $msgid ) = MakeMessageID( );
  local( $tmp_auth, $tmp_mail, $tmp_subj, $tmp_mess );
  local( $msg ) = "";

  if( $previewtext || $posttext ){
    $tmp_auth = HTMLescape( $author );
    $tmp_mail = HTMLescape( $mail_address );
    $tmp_subj = HTMLescape( $subject );
    $tmp_subj = URL2tag( $tmp_subj );
    $tmp_msg = HTMLescape( $message );
    $tmp_msg = URL2tag( $tmp_msg );
  } else {
    $tmp_auth = $author;
    $tmp_mail = $mail_address;
    $tmp_subj = $subject;
    $tmp_msg = $message;
  }

  $msg .= "<div class=\"article\" id=\"$msgid\">\n";  ## </div> is at 20 lines below.
  $msg .= "<div class=\"atcl-head\" id=\"$msgid-head\">\n";  ## </div> is at 29 lines below.
  if( $USE_REV ){
    $msg .= "<STRONG>Message-Id: </STRONG>&lt;<a name=\"$msgid\" href=\"$realrev#$msgid\">$msgid</a>&gt;<br>\n";
  }else{
    $msg .= "<STRONG>Message-Id: </STRONG>&lt;<a name=\"$msgid\" href=\"$reallog#$msgid\">$msgid</a>&gt;<br>\n";
  }

  $msg .= "<STRONG>Date: </STRONG>$dname $mname $mday $hh:$mm:$ss $TZNAME $y<br>\n";
  $msg .= "<STRONG>From: </STRONG>";
  if( $tmp_mail =~ /^https?:\/\/[a-zA-Z0-9;\/\?:@&=+$,---_.!~*'()<>#%"{}|\\^\[\]`]+$/ ){
    $msg .= "<a href=\"$tmp_mail\">$tmp_auth</a>";
  }else{
    $msg .= "$tmp_auth";
    if( !$tmp_mail eq '' ) {
      $msg .= " &lt;$tmp_mail&gt;";
    }
  }
  $msg .= "<br>\n";
  $msg .= "<STRONG>Subject: </STRONG>$tmp_subj<br>\n";
  $msg .= "<br></div>\n\n";
  $msg .= "<div class=\"atcl-body\" id=\"$msgid-body\">\n";
  if( $previewtext || $posttext ){
    $msg .= "<p>\n";
    $tmp_msg =~ s/\r\n/<br>/g;
    $tmp_msg =~ tr/\r/<br>/;
    $tmp_msg =~ tr/\n/<br>/;
    $tmp_msg =~ tr/<br>/<br>\n/;
    $msg .= "$tmp_msg\n";
    $msg .= "</p>\n";
  } else {
    if( $USE_FORMULA ){
      for( $i=1; $tmp_msg =~ s/`([^`]*)`/mkFormulaTag($msgid,$i,$1)/e; $i++ ){}
      if( $i != 1 ){
        $form_tag .= '<div id="' . $msgid . '-button">';
        $form_tag .= '<a class="button" onClick="javascript:showArticleInMathML(';
        $form_tag .= "\'$msgid\', \'0\'";
        $form_tag .= ')">[showArticle in MathML]</a></div>';
        $tmp_msg .= $form_tag;
      }
    }
    $msg .= "$tmp_msg\n";
  }
  $msg .= "</div></div>\n";

  $msg =~ s/\r\n/\n/g;
  $msg =~ tr/\r/\n/;

  &jcode'convert( *msg, 'euc' );
  return $msg;
}

#####
# Input message

sub InputMessage {
  local( $tmptxt ) = "";

  $tmptxt = LogHead( "$TITLE: Post Message", 'cgi' );

  if( !( $preview ) & !( $previewtext ) ){
    if( open( HTP, "$HOWTO_FILE" ) ){
      $tmptxt .= "<hr>\n\n";
      while( <HTP> ){
        &jcode'convert( *_, 'euc' );
        $tmptxt .= $_;
      }
      close( HTP );
      $tmptxt .= "\n";
    }
  }

  &jcode'convert( *tmptxt, $KANJI_CODE );
  print $tmptxt;

  $tmptxt = "";
  if( $preview || $previewtext ){
    local( $msg, $error );

    if( $author eq '' || $subject eq '' || $message eq '' ) {
      $tmptxt .= "<hr>\n\n<h2>Error</h2>";
      $tmptxt .= "<p>Name, subject, and message is necessary.</p>";
      $tmptxt .= "<form method=\"POST\" action=\"$BBS_CGI\">\n\n";
    }elsif( $WEBLINT ){
      local( $wlerror ) = 0;
      $tmptxt .= ResultWeblint();
      $tmptxt .= "<form method=\"POST\" action=\"$BBS_CGI\">\n\n";
    } else {
      $tmptxt .= "<form method=\"POST\" action=\"$BBS_CGI\">\n\n";

      $msg = ParseMessage();

      $tmptxt .= "<hr>\n\n";
      $tmptxt .= "<h2>Preview</h2>\n\n";
      if( ( $error = CheckTag( $msg ) )
          || ( $NEED_NON_ASCII 
               && !$esp 
               && ( $message =~ /^[\x01-\x7f]*$/ || $subject =~ /^[\x01-\x7f]*$/ ) ) ){
          $tmptxt .= "<p>Error: Your message contains a bad tag: $error.</p>\n\n";
      } else {
        $tmptxt .= "$msg";
        $tmptxt .= "\n<p>\n";
        if( $preview ){
          $tmptxt .= "<input type=\"submit\" value=\"Post HTML\" name=\"post\">\n";
        }else{
          $tmptxt .= "<input type=\"submit\" value=\"Post TEXT\" name=\"posttext\">\n";
        }
        $tmptxt .= "</p>\n\n";
      }
    }
  } else {
    $tmptxt .= "<form method=\"POST\" action=\"$BBS_CGI\">\n\n";
  }

  $tmptxt .= "<hr>\n\n";
  $tmptxt .= "<h2>Input Form</h2>\n\n";

  &jcode'convert( *tmptxt, $KANJI_CODE );
  print $tmptxt;

  PrintInputForm( );

  $tmptxt = HTMLfoot();

  &jcode'convert( *tmptxt, $KANJI_CODE );
  print $tmptxt;
}

#####
# fatal Error in HTML

sub FatalError {
  local( $errmsg ) = @_;
  local( $tmptxt ) = "";

  Unlock( $LOCK_FILE );

  $tmptxt = LogHead( "$TITLE: Error", 'cgi' );
  $tmptxt .= "<hr><h2>Error</h2><p>$errmsg</p>\n";
  $tmptxt .= HTMLfoot();

  &jcode'convert( *tmptxt, $KANJI_CODE );
  print $tmptxt;
  exit( 0 );
}

#####
# make header part of $LOG.html

sub LogHead {
  local( $title, $pos ) = @_;
  local( $tmp ) = "";

  $tmp = HTMLhead( $title );
  $tmp .= HTMLlinks( $pos );

  &jcode'convert( *tmp, 'euc' );
  return $tmp;
}

#####
# Message from Admin

sub AdmMsg {
  local( $tmp ) = "";

  $tmp .= "<hr>\n";
  if( open( AM, $ADMMSG_FILE ) ){
    while( <AM> ){
      $tmp .= $_;
    }
    close( AM );
    return $tmp;
  }
}

#####
# Post Log

sub PostLog {
  if( open( BBSLOG, ">>$REAL_POSTLOG" )) {
    local( $tmptxt ) = "";

    local( $tmp_author ) = $author;
    $tmp_author =~ tr/\r\n/ /;
    $tmp_author =~ s/<[^>]*>//g;
    $tmp_author = HTMLescape( $tmp_author );

    local( $tmp_mail_address ) = $mail_address;
    $tmp_mail_address =~ tr/\r\n/ /;
    $tmp_mail_address =~ s/<[^>]*>//g;
    $tmp_mail_address = HTMLescape( $tmp_mail_address );

    local( $tmp_subject ) = $subject;
    $tmp_subject =~ tr/\r\n/ /;
    $tmp_subject =~ s/<[^>]*>//g;
    $tmp_subject = HTMLescape( $tmp_subject );

    $tmp_remote = $ENV{'REMOTE_HOST'};
    $tmp_remote = $ENV{'REMOTE_ADDR'} unless $tmp_remote;

    $tmptxt .= "$msgid $tmp_remote \"$ENV{'HTTP_USER_AGENT'}\" ";
    $tmptxt .= "\"$tmp_author\" \"$tmp_mail_address\" \"$tmp_subject\"\n";

    &jcode'convert( *tmptxt, $KANJI_CODE );
    print BBSLOG $tmptxt;
    close( BBSLOG );
    chmod 0666, $realpostlog;
  }
}

#####
# Basic Checks

sub CheckMessage {
  local( $msg ) = @_;

  ## check tags
  if( $error = CheckTag( $msg ) ){
    FatalError( "Your message contains a bad tag: $error." );
  }

  ## check file lock
  for( $i = 0; !Lock( $LOCK_FILE ); $i++ ){
    if( $i > 5 ){
      FatalError( "Temporary file is locked. Try again, later. $LOCK_FILE" );
    }
    select undef, undef, undef, 0.1;
  }

  ## check duplicate posting
  if( !open( LOG, "$REAL_LOG" )) {
    FatalError( "Can not open log file: $REAL_LOG." );
  }
  while( <LOG> ){
    if( $_ =~ /$TOP/ ){ last; }
  }
  while( <LOG> ){
    if( $_ =~ /$BDR/ ){ last; }
    $_ =~ s/<STRONG>Message-Id: <\/STRONG>.*\n//;
    $_ =~ s/<STRONG>Date: <\/STRONG>.*\n//;
    $lastmsg .= $_;
  }
  $tmptxt = "\n$msg\n<hr>\n";
  $tmptxt =~ s/<STRONG>Message-Id: <\/STRONG>.*\n//m;
  $tmptxt =~ s/<STRONG>Date: <\/STRONG>.*\n//m;
  if( $lastmsg eq $tmptxt ){
    FatalError( "Duplicated Messages." );
  }

  ## check files
  if( !open( TMP, ">$TMP_FILE" ) ){
    FatalError( "Can not write temporary file: $TMP_FILE" );
  }
  if( !open( LOG, "$REAL_LOG" ) ){
    FatalError( "Can not open log file: $REAL_LOG." );
  }

  if( !open( IDX, ">$INDEX_FILE" ) ){
    FatalError( "Can not open index file: $INDEX_FILE." );
  }

  if( $USE_REV ){
    if( !open( REV, "$REAL_REV" ) ){
      FatalError( "Can not open reversed log file: $REAL_REV." );
    }
    if( !open( TMP_REV, ">$TMP_REV_FILE" ) ){
      FatalError( "Can not write temporary rev_file: $TMP_REV_FILE" );
    }
  }

}

#####
# Post Message

sub PostMessage {
  local( $tmptxt, $tmplog, $error, $lastmsg );
  if( $USE_REV ){
    local( $tmprev );
  }

  if( $author eq '' || $subject eq '' || $message eq '' ) {
    FatalError( "Name, subject, and message is necessary." );
  } elsif( ( $error = CheckTag( $msg ) )
          || ( $NEED_NON_ASCII 
               && !$esp 
               && ( $message =~ /^[\x01-\x7f]*$/ || $subject =~ /^[\x01-\x7f]*$/ ) ) ){
    FatalError( "Your message contains a bad tag." );
  }

  $msg = ParseMessage();

  if( $WEBLINT ){
    local( $wlerror ) = 0;
    $tmptxt .= ResultWeblint();

    if( $wlerror != 0 ){
      local( $preview ) = "1";
      InputMessage();
      return;
    }
  }

  CheckMessage( $msg );

  ## make tmpfile for new $LOG.html
  $tmptxt = LogHead( $TITLE, 'spool' );
  $tmptxt .= SearchForm( $key, 'spool' );
  &jcode'convert( *tmptxt, $KANJI_CODE );
  print TMP $tmptxt;
  print IDX $tmptxt;
  if( $USE_REV ){
    print TMP_REV $tmptxt;
  }

  $tmptxt = AdmMsg( );
  $tmptxt .= "<hr>\n$TOP\n";
  &jcode'convert( *tmptxt, $KANJI_CODE );
  print TMP $tmptxt;
  if( $USE_REV ){
    print TMP_REV $tmptxt;
  }

  $tmptxt = "\n$msg\n<hr>\n$BDR\n";

  ## get old messages
  while( <LOG> ) {
    if( $_ =~ /$TOP/ ) { last; }
  }
  while( <LOG> ) {
    if( $_ =~ /$END/ ) { last; }
    $tmplog .= $_;
  }
  close( LOG );

  if( $USE_REV ){
    while( <REV> ) {
      if( $_ =~ /$TOP/ ) { last; }
    }
    while( <REV> ) {
      if( $_ =~ /$END/ ) { last; }
      $tmprev .= $_;
    }
    close( REV );
  }

  &jcode'convert( *tmptxt, $KANJI_CODE );
  &jcode'convert( *tmplog, $KANJI_CODE );
  print TMP  $tmptxt;
  print TMP  $tmplog;

  &jcode'convert( *GOTOLAST, $KANJI_CODE );
  if( $oldlog ne '' ){
    print TMP "<div><a href=\"$oldlog\">$GOTOLAST</a></div>\n";
  }
  if( $USE_REV ){
    if( $oldlog ne '' ){
      print TMP_REV "<div><a href=\"$oldrev\">$GOTOLAST</a></div>\n<hr>\n";
    }
    print TMP_REV $tmprev;
    print TMP_REV $tmptxt;
  }

  ## make index
  $tmptxt = "\n<hr>\n\n<h2>Last Modified at...</h2>\n\n<pre>\n";
  $tmptxt .= "<a href=\"$LOG_HTML\">$TITLE</a>       <a href=\"$reallog?$y$m$d$hh$mm$ss\">$y/$m/$d $hh:$mm:$ss</a>\n";
  $tmptxt .= "</pre>\n\n";
  &jcode'convert( *tmptxt, $KANJI_CODE );
  print IDX $tmptxt;

  ## print footer
  $tmptxt = "$END\n\n";
  $tmptxt .= "<div><a href=\"#top\">$GOTOTOP</a></div>\n";
  if( -f "$TOP_DIR/$SPOOL/$OLD_HTML" ) {
    $tmptxt .= "<a href=\"$OLD_HTML\">$READOLD</a>\n";
  }
  &jcode'convert( *tmptxt, $KANJI_CODE );
  print TMP $tmptxt;
  if( $USE_REV ){
    print TMP_REV $tmptxt;
  }

  $tmptxt = HTMLfoot();
  &jcode'convert( *tmptxt, $KANJI_CODE );
  print TMP $tmptxt;
  print IDX $tmptxt;
  if( $USE_REV ){
    print TMP_REV $tmptxt;
    close( TMP_REV );
    if( !rename( $TMP_REV_FILE, $REAL_REV ) ){
      FatalError( "Can not rename temp_rev file to log file.<br>$TMP_REV_FILE<br>$REAL_REV" );
    }
    chmod 0666, $REAL_REV;
  }

  close( TMP );
  close( IDX );

  if( !rename( $TMP_FILE, $REAL_LOG ) ){
    FatalError( "Can not rename temporary file to log file.<br>$TMP_FILE<br>$REAL_LOG" );
  }
  chmod 0666, $REAL_LOG;

  ## for USE_COPY
  if( $USE_COPY ){
    copy( $REAL_LOG, $LOG_FILE );
    if( $USE_REV ){
      copy( $REAL_REV, $LOG_REV_FILE );
    }
  }
  ##

  if( $SENDMAIL ){
    NoticeMail();
  }
  PostLog( );
  Unlock( $LOCK_FILE );

  $tmptxt = LogHead( "$TITLE: Message Accepted", 'cgi' );

  $tmptxt .= "<hr>\n\n";
  $tmptxt .= "<h2>Thank you for posting a message!\n</h2>\n";
  $tmptxt .= "$msg\n";
  $tmptxt .= HTMLfoot();

  &jcode'convert( *tmptxt, $KANJI_CODE );
  print $tmptxt;  
}

#####
# Send notice by mail
sub NoticeMail {
  local( $mailbody, $mailhead, $mailsubj, $mailfrom );

  if( !open( SM, "| $SENDMAIL -t -i -f $MAILADD" ) ){
    return;
  }

  require "$LIB_DIR/mimew.pl";

  $mailsubj = &mimeencode( $MAILSUBJ );
  $mailfrom = &mimeencode( "$MAILADD ($TITLE)" );
  $mailhead = <<END;
To: $MAILTO
From: $mailfrom
Subject: $mailsubj

END

  $mailbody = <<END;
new message is coming:
  From $author
  Subject $subject

See:
  $FULLURL
END

  &jcode'convert( *mailbody, 'jis' );

  print SM $mailhead;
  print SM $mailbody;

  close( SM );
}

#####
# Create or rewrite old log

sub OldLog {
  local( $tmptxt, $tmpline, $ser ) = "";

  if( ! -f $OLDLOG_FILE ){
    ## Create old log file
    if( !open( OLD, ">$OLDLOG_FILE" ) ){
      FatalError( "Can not open $OLDLOG_FILE." );
    }
    $tmptxt .= LogHead( "All Logs of $TITLE", 'spool' );
    $tmptxt .= "<hr>\n\n<ul>\n$TOP\n";
    $tmptxt .= "$END\n</ul>\n\n";
    $tmptxt .= HTMLfoot();

    &jcode'convert( *tmptxt, $KANJI_CODE );
    print OLD $tmptxt;
    close( OLD );
    chmod 0666, $OLDLOG_FILE;
  }

  if( !open( OLD, $OLDLOG_FILE ) ){
    FatalError( "Can not open $OLDLOG_FILE." );
  }

  $tmptxt = "";
  while( $_ = <OLD> ){
    $tmptxt .= $_;
    if( $_ =~ /$TOP/ ){ last; }
  }

  if( $SIZE ){
    $ser = "No. $serial";
  } else {
    $ser = "in $y/$m";
  }
  if( $USE_REV ){
    $tmpline .= "<li><a href=\"$realrev\">$TITLE $ser</a> (<a href=\"$reallog\">reverse order</a>)\n";
  }else{
    $tmpline .= "<li><a href=\"$reallog\">$TITLE $ser</a>\n";
  }
  &jcode'convert( *tmpline, $KANJI_CODE );
  $tmptxt .= $tmpline;

  while( $_ = <OLD> ){
    $tmptxt .= $_;
    if( $_ =~ /$END/ ){ last; }
  }
  close( OLD );
  $tmptxt .= "</ul>\n\n";
  $tmpline = HTMLfoot();
  &jcode'convert( *tmpline, $KANJI_CODE );
  $tmptxt .= $tmpline;

  if( !open( OLD, ">$OLDLOG_FILE" ) ){
    FatalError( "Can not open $OLDLOG_FILE." );
  }
  &jcode'convert( $tmptxt, $KANJI_CODE );
  print OLD $tmptxt;
  close( OLD );
}

#####
# Get Serial Number of log file

sub GetSerial {
  ## for compatibility for old version
  if( $SIZE ){
    $SER = '.seq';
  } else {
    $SER = '.month';
  }
  if( !-f "$TOPDIR/$SPOOL/$SER" ){
    if( open( SEQ, ">$TOPDIR/$SPOOL/$SER" ) ){
      print SEQ "----\n";
      close( SEQ );
    }
    if( !open( SEQ, "$TOPDIR/$SPOOL/$SER" ) ){
      FatalError( "cannot open $SER file." );
    }
  } else {
    if( !open( SEQ, "$TOPDIR/$SPOOL/$SER" ) ){
      FatalError( "cannot open $SER file." );
    }
  }
  chmod 0666, "$TOPDIR/$SPOOL/$SER";

  $serial = <SEQ>;
  chop( $serial );
  close( SEQ );
  return( $serial );
}


#####
# Create New $SER File

sub NewSeq {
  local( $new ) = @_;

  if( !open( SEQ, ">$TOPDIR/$SPOOL/$SER" ) ){
    FatalError( "Cannot write to $SER file: $new" );
  }
  print SEQ "$new\n";
  close( SEQ );
}

#####
# Create Log File

sub CreateLogFile {
  if( ! -f $INDEX_FILE ){
    if( !open( IDX, ">$INDEX_FILE" ) ){
      FatalError( "Can not create index file: $INDEX_FILE." );
    }
  }
  close( IDX );

  $serial = GetSerial();

  ## check real log file
  if( $SIZE ){
    if( $serial eq '----' ){
      $newser = '0000';
      NewLog( $newser );
    } else {
      $reallog = $LOG . $serial . '.html';  # LOGyyyymm.html
      $REAL_LOG = "$TOPDIR/$SPOOL/$reallog";
      if( -s $REAL_LOG > $SIZE * 1024 ){    # size over?
        $newser = sprintf( "%04d", $serial + 1 );
        NewLog( $newser );
      }
    }
  } else {
    $newser = $y . $m;
    if( ! -f "$TOPDIR/$SPOOL/$LOG$newser.html" ){
      NewLog( $newser );
    } else {
      $serial = $newser;
      NewSeq( $newser );
    }
  }
  $reallog = $LOG . $serial . '.html';  # LOGyyyymm.html
  $REAL_LOG = "$TOPDIR/$SPOOL/$reallog";
  $realpostlog = $POST_LOG . $serial;  # POST_LOGyyyymm.html
  $REAL_POSTLOG = "$TOPDIR/$SPOOL/$realpostlog";
  if( $USE_REV ){
    $realrev = $REV . $serial . '.html';  # REVyyyymm.html
    $REAL_REV = "$TOPDIR/$SPOOL/$realrev";
  }
}

#####
# Make Log File

sub NewLog {
  local( $newser ) = @_;
  $newlog = $LOG . $newser . '.html';
  $NEW_LOG = "$TOPDIR/$SPOOL/$newlog";
  $newpostlog = $POST_LOG . $newser;
  $NEW_POSTLOG = "$TOPDIR/$SPOOL/$newpostlog";
  if( $USE_REV ){
    $newrev = $REV . $newser . '.html';
    $NEW_REV = "$TOPDIR/$SPOOL/$newrev";
  }

  if( !open( LOG, ">$NEW_LOG" ) ){
    FatalError( "Can not create new log file: $NEW_LOG." );
  }
  if( !open( POSTLOG, ">$NEW_POSTLOG" ) ){
    FatalError( "Can not create new postlog file: $NEW_POSTLOG." );
  }
  if( $USE_REV ){
    if( !open( LOG_REV, ">$NEW_REV" ) ){
      FatalError( "Can not create new reverse log file: $NEW_REV." );
    }
  }
  
  close( LOG );
  unlink( $LOG_FILE );
  close( POSTLOG );
  unlink( $POSTLOG_FILE );
  if( $USE_REV ){
    close( LOG_REV );
    unlink( $LOG_REV_FILE );
  }

  if( $USE_COPY ){
    ## for USE_COPY
    if( !open( LOG, ">$LOG_FILE" ) ){
      FatalError( "Can not create log file: $LOG_FILE." );
    }
    if( !open( POSTLOG, ">$POSTLOG_FILE" ) ){
      FatalError( "Can not create postlog file: $POSTLOG_FILE." );
    }
    close( LOG );
    close( POSTLOG );
    if( $USE_REV ){
      if( !open( LOG_REV, ">$LOG_REV_FILE" ) ){
        FatalError( "Can not create log file: $LOG_REV_FILE." );
      }
      close( LOG_REV );
    }
    ##
  }else{
    if( !symlink( $NEW_LOG, $LOG_FILE ) ){
      FatalError( "Can not make symlink: $NEW_LOG from $LOG_FILE: $!" );
    }
    if( !symlink( $NEW_POSTLOG, $POSTLOG_FILE ) ){
      FatalError( "Can not make symlink: $NEW_POSTLOG from $POSTLOG_FILE: $!" );
    }
    if( $USE_REV ){
      if( !symlink( $NEW_REV, $LOG_REV_FILE ) ){
        FatalError( "Can not make symlink: $NEW_REV from $LOG_REV_FILE: $!" );
      }
    }
  }

  if( $serial ne '----' ){
    $oldlog = $LOG . $serial . '.html';
    $oldpostlog = $POST_LOG . $serial;
    if( $USE_REV ){
      $oldrev = $REV . $serial . '.html';
    }
  }
  $serial = $newser;
  $reallog = $LOG . $serial . '.html';  # LOGyyyymm.html
  $REAL_LOG = "$TOPDIR/$SPOOL/$reallog";
  $realpostlog = $POST_LOG . $serial;   # POST_LOGyyyymm
  $REAL_POSTLOG = "$TOPDIR/$SPOOL/$realpostlog";
  if( $USE_REV ){
    $realrev = $REV . $serial . '.html';  # REVyyyymm.html
    $REAL_REV = "$TOPDIR/$SPOOL/$realrev";
  }

  OldLog();
  NewSeq( $newser );
}

sub HTMLdigest {
  local( $tmptxt, $file );

  if( $digest ne "" ){
    $POSTLOG_FILE .= "$digest";
  }
  if( $USE_REV  == 1 ){
    $file = $REV;
  } else {
    $file = $LOG;
  }

  if( ! open( PLF, "$POSTLOG_FILE" ) ){
    FatalError( "Cannot open post_log file: $POSTLOG_FILE." );
  }

  $tmptxt = LogHead( "$TITLE: digest$digest", 'cgi' );
  $tmptxt .= "<hr>\n\n<dl>\n";
  while( <PLF> ){
    chop( $_ );
    $_ =~ s/<[^>]*>//g;
    $_ =~ s/^(....)(..)(..)(..)(..)(..)[^\t]* ([^\t]*) \"([^\t]*)\" \"([^\t]*)\" \"([^\t]*)\" \"(.*)\"$/<dt><a href=\"$SPOOL\/$file$digest.html#$1$2$3$4$5$6\@$DOGTAG\">$11<\/a><dd>$9 \($1\/$2\/$3 $4:$5:$6\)\n/;
    $tmptxt .= $_;
  }
  $tmptxt .= "</dl>\n";
  $tmptxt .= HTMLfoot();
  &jcode'convert( $tmptxt, $KANJI_CODE );
  print $tmptxt;
}

sub GetAtclHdr {
  local( $tgt, $lnum ) = @_;
  local( $tmpdate, $tmpfrom, $tmpsubj );

  for( $lnum++ ; <$tgt>; $lnum++ ){
    if( /^$/ ){
      return( $lnum, $tmpdate, $tmpfrom, $tmpsubj );
    } elsif ( /^<(STRONG|b)>Date: <\/(STRONG|b)>(.*)<br>$/ ){
      $tmpdate = $3
    } elsif ( /^<(STRONG|b)>From: <\/(STRONG|b)>(.*)<br>$/ ){
      $tmpfrom = $3;
    } elsif ( /^<(STRONG|b)>Subject: <\/(STRONG|b)>(.*)<br>$/ ){
      $tmpsubj = $3;
    }
  }
  return( $lnum, $tmpdate, $tmpfrom, $tmpsubj );
}

sub SearchForm {
  local( $key, $pos ) = @_;
  local( $tmp ) = "";

  if( $pos eq 'spool' ){
    $tmp = "<form action=\"../$BBS_CGI\" method=\"get\">";
  } else {
    $tmp = "<form action=\"./$BBS_CGI\" method=\"get\">";
  }
  $tmp .= "<div style=\"text-align:right\">\n";
  $tmp .= "$TITLE Search: <input type=\"text\" name=\"search\" value=\"$key\">\n";
  $tmp .= "<input type=\"submit\" value=\"Search\">\n</div></form>\n";
  &jcode'convert( $tmp, 'euc' );
  return( $tmp );
}

sub Search {
  $GREP = '/usr/local/gnu/bin/grep';
  local( $key ) = @_;
  local( $tmp, $files, $grepcmd, $prev_msgid, $prev_file ) = "", "", "";
  local( $target_line, $target_lnum, $msg_id, $prev_id ) = "", 0, "", "";
  local( $tmpdate, $tmpsubj, $tmpfrom ) = "", "", "";
  local( $found, $first_id ) = 0, 0;

  ## check arguments
  if( $key eq "" ){
    FatalError( 'Please input any keywords...' );
  }
  $_ = $key;
  if( /[`"';]/ ){
    FatalError( 'Don\'t use \', `, \", and ;.' );
  }

  ## make grep command
  if( $USE_REV = 1 ){
    $files = $REV;
  } else {
    $files = $LOG;
  }
  $grepcmd = "$GREP -ni '$key' $TOPDIR/$SPOOL/$files\_[0-9]*.html $TOPDIR/$SPOOL/$files\[0-9]*.html /dev/null";
  if( ! open( GRP, "$grepcmd |" ) ){
    FatalError( "Can't execute grep command:<br>$grepcmd" );
  }

  $tmp = LogHead( "$TITLE Search Result", 'cgi' );
  $tmp .= SearchForm( $key, 'cgi' );
  $tmp .= "<h2>Search Result of $TITLE</h2>";
  $tmp .= "<hr>\n\n"; 

  ## read result of grep
  while( <GRP> ){
    if( /^([^:]*):([0-9]*):/ ){
      $filename = $1;
      $line = $2;
    } else {
      FatalError( "grep error." );
    }
    $prev_id = $msg_id;

    if( $filename ne $prev_file ){
      $prev_file = $filename;
      if( ! open( TARGET, "$filename" ) ){
        FatalError( "Cannot open search target: $filename" );
      }

      ## forward to the 1st Message-Id
      for( $target_lnum = 1; <TARGET>; $target_lnum++ ){
        if( /^<(STRONG|b)>Message-Id: <\/(STRONG|b)>&lt;<a name=\"([^\"]*)\"/ ){
          $msg_id = $3;
          $first_id = $target_lnum;
          ( $target_lnum, $tmpdate, $tmpfrom, $tmpsubj ) = GetAtclHdr( 'TARGET', $target_lnum );
          last;
        }
      }
      if( $_ eq "" ){
        ## when log file doesn't include "Message-Id"
        FatalError( "no \"Message-Id\"s in $filename<br>$_" );
      }
    }  ## end of NEW log*.html

    if( $first_id > $line ){
      ## hit on the header. ignored
      next;
    }

    ## seek on the log file
    for( $target_lnum++; <TARGET>; $target_lnum++ ){
      if( $target_lnum >= $line ){
        ## article was found
        last;
      }
      if( /^<(STRONG|b)>Message-Id: <\/(STRONG|b)>&lt;<a name=\"([^"]*)\"/ ){
        $msg_id = $3;
        ( $target_lnum, $tmpdate, $tmpfrom, $tmpsubj ) = GetAtclHdr( 'TARGET', $target_lnum );
      }
    }
    if( $prev_id eq $msg_id ){
      ## same article
      next;
    }

    if( $found == 0 ){
      $found = 1;
      $tmp .= "<ul>\n";
      &jcode'convert( *tmp, $KANJI_CODE );
      print $tmp;
    }

    $tmpfrom =~ s/<[^>]*>//g;
    $tmpsubj =~ s/<[^>]*>//g;
    $tmp = "  <li><STRONG>[$tmpfrom]</STRONG> ($tmpdate)<br>\n";
    $filename =~ /^.*\/([^\/]*.html)$/;
    $tmp .= "    <a href=\"$SPOOL/$1\#$msg_id\">$tmpsubj</a>\n";
    &jcode'convert( *tmp, $KANJI_CODE );
    print $tmp;
  }

  ## found any articles
  if( $found == 0 ){
    $tmp .= "no hits..."; 
  } else {
    $tmp = "</ul>"; 
  }
  $tmp .= "\n\n";
  $tmp .= HTMLfoot();
  &jcode'convert( *tmp, $KANJI_CODE );
  print $tmp;
}
1;
